import { singular } from 'pluralize'
import {
  IGQLType,
  IGQLField,
  ISDL,
  capitalize,
  toposort,
  isTypeIdentifier,
  camelCase,
} from 'prisma-datamodel'
import { Normalizer } from './normalizer'

export default class ModelNameNormalizer extends Normalizer {
  public normalize(model: ISDL) {
    // We need to sort types according to topological order for name normalization.
    // Otherwise embedded type naming might break as embedded types depend on
    // their parent type.
    this.normalizeTypes(toposort(model.types), model)
  }

  protected assignName(obj: IGQLType | IGQLField, newName: string) {
    if (obj.databaseName === null) {
      // If name was already changed, we don't touch it.
      if (newName !== obj.name) {
        // If name is already conforming to prisma, skip.
        obj.databaseName = obj.name
        obj.name = newName
      }
    }
  }

  protected getNormalizedTypeName(name: string, model: ISDL) {
    if (name.toUpperCase() === name) {
      return name
    }

    // Replace all special chars.
    let normalizedName = name.replace(/[^a-z0-9_]+/gi, '')

    normalizedName = capitalize(camelCase(singular(normalizedName)))

    // if there is a naming conflict with a known scalar type, use the default name
    if (isTypeIdentifier(normalizedName) || isTypeIdentifier(singular(name))) {
      return name
    }

    // if there is already a table in the database with the exact name we're generating - let's just not do it
    if (model.types.some(t => t.name === normalizedName)) {
      return name
    }

    return normalizedName
  }

  protected normalizeType(
    type: IGQLType,
    model: ISDL,
    forceNoRename: boolean = false,
  ) {
    if (!forceNoRename) {
      this.assignName(type, this.getNormalizedTypeName(type.name, model))
    }

    super.normalizeType(type, model)
  }

  protected getNormalizedFieldName(
    name: string,
    parentType: IGQLType,
    field: IGQLField,
  ) {
    // For all-uppercase field names, we do not normalize.
    if (name.toUpperCase() === name) {
      return name.toLowerCase()
    }

    // Trim _id from related fields.
    if (typeof field.type !== 'string' && name.toLowerCase().endsWith('_id')) {
      name = name.substring(0, name.length - 3)
    }

    // Replace all special chars.
    name = name.replace(/[^a-z0-9_]+/gi, '')

    // Follow prisma conventions.
    const normalizedName = camelCase(name)

    // If there is already a field in this type for the normalized name, don't rename.
    const conflictingField = parentType.fields.find(
      f => f.name === normalizedName && f !== field,
    )
    if (conflictingField) {
      return null
    }

    return normalizedName
  }

  protected normalizeField(
    field: IGQLField,
    parentType: IGQLType,
    parentModel: ISDL,
  ) {
    // Make field names pretty
    if (!parentType.isEnum && !field.isId) {
      let normalizedName = this.getNormalizedFieldName(
        field.name,
        parentType,
        field,
      )
      if (normalizedName === null) {
        field.comments.push({
          isError: false,
          text:
            'Field name normalization failed because of a conflicting field name.',
        })
      } else {
        this.assignName(field, normalizedName)
      }
    }

    // Make embedded type names pretty
    if (typeof field.type !== 'string' && field.type.isEmbedded) {
      if (!field.type.databaseName) field.type.databaseName = field.type.name

      field.type.name = parentType.name + capitalize(singular(field.name))
    }
  }
}
